/* Copyright 2016 Jim Snow
 * GPL version 2 */

#include <wiringPi.h>
#include "mcp3004.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include <linux/soundcard.h>
#include <stdint.h>
//#include <alsa/asoundlib.h>

#include "midi.h"

#define max(a, b) (a > b ? a : b)

#define SREG_CLK 0
#define SREG_DATA 1
#define ADC_CS 2

#define UP 2
#define DOWN 3

//#define adc_cs_high digitalWrite(ADC_CS, HIGH)
//#define adc_cs_low digitalWrite(ADC_CS, LOW)

//#define TICK 400000
#define TICK 0
#define BASE 100
#define SPI_CHAN 0
#define ADCS 20

int midifd;
char *dev = "/dev/midi1";
int enable_pan = 0;
int instrument_num;

//static int midi_client;
//snd_seq_t *seq;

char * names[ADCS*8] = {
	"(1,4)","(25,96)","(17,64)","(4,15)","(35,128)","(5,18)","(9,32)","(7,24)","(75,256)","(19,64)","(3,10)","(5,16)","(21,64)","(1,3)","(27,80)","(11,32)","(25,72)","(45,128)","(35,96)","(3,8)","(25,64)","(2,5)","(13,32)","(105,256)","(5,12)","(27,64)","(7,16)","(4,9)","(9,20)","(15,32)","(63,128)", "none",
	"(1,2)","(25,48)","(17,32)","(8,15)","(35,64)","(5,9)","(9,16)","(7,12)","(75,128)","(19,32)","(3,5)","(5,8)","(21,32)","(2,3)","(27,40)","(11,16)","(25,36)","(45,64)","(35,48)","(3,4)","(25,32)","(4,5)","(13,16)","(105,128)","(5,6)","(27,32)","(7,8)","(8,9)","(9,10)","(15,16)","(63,64)", "none",
	"(1,1)","(25,24)","(17,16)","(16,15)","(35,32)","(10,9)","(9,8)","(7,6)","(75,64)","(19,16)","(6,5)","(5,4)","(21,16)","(4,3)","(27,20)","(11,8)","(25,18)","(45,32)","(35,24)","(3,2)","(25,16)","(8,5)","(13,8)","(105,64)","(5,3)","(27,16)","(7,4)","(16,9)","(9,5)","(15,8)","(63,32)", "none",
	"(2,1)","(25,12)","(17,8)","(32,15)","(35,16)","(20,9)","(9,4)","(7,3)","(75,32)","(19,8)","(12,5)","(5,2)","(21,8)","(8,3)","(27,10)","(11,4)","(25,9)","(45,16)","(35,12)","(3,1)","(25,8)","(16,5)","(13,4)","(105,32)","(10,3)","(27,8)","(7,2)","(32,9)","(18,5)","(15,4)","(63,16)", "none",
	"(4,1)","(25,6)","(17,4)","(64,15)","(35,8)","(40,9)","(9,2)","(14,3)","(75,16)","(19,4)","(24,5)","(5,1)","(21,4)","(16,3)","(27,5)","(11,2)","(50,9)","(45,8)","(35,6)","(6,1)","(25,4)","(32,5)","(13,2)","(105,16)","(20,3)","(27,4)","(7,1)","(64,9)","(36,5)","(15,2)","(63,8)", "(8,1)"};

struct key {
  uint16_t ratio_a;
  uint16_t ratio_b;
  char *name;
  uint64_t last_send;
  int threshold_min;
  int threshold_max;
  int midinote;
  uint8_t pitchbend_msb;
  uint8_t pitchbend_lsb;
  uint8_t channel;
  uint8_t pan;
  uint16_t maxval;
  uint16_t minval;
};

struct key keys[ADCS*8];

/* frequency of 1/1 in hz */
// double key_freq = 512.0;    /* C in C256 */
double key_freq = 391.995*1.0; /* G in A440 */
double note0_freq = 8.1757989156;

double curve_exponent = 1.0; // 0.5;

/* proteus defaults to 2 */
double pb_semitones = 2.0;

#include "min_calibration.h"
#include "max_calibration.h"

void noteinit () {
  double semitone = pow(2.0, 1.0/12.0);
  double pb_increment = pow(2.0, pb_semitones/(12.0 * 8192.0));
  int i;

  for (i = 0; i < ADCS*8; i++) {
    int a,b;
    double freq;
    uint8_t note = 0;
    uint16_t pb = 0;

    if (sscanf(names[i], "(%d,%d)", &a, &b) != 2) {
       keys[i] = (struct key) {
         .ratio_a = 0,
         .ratio_b = 0,
         .name = names[i],
         .channel = CHANNEL_NONE,
         .minval = 1023,
         .maxval = 0
       };
       printf("skipping key %d\n", i);
       continue;
    }
    freq = (key_freq * (double)a) / (double)b;

    while (note0_freq * pow(semitone, note) < freq && note < 128) {
      note++;
    }

    note--;

    while (note0_freq * pow(semitone, note) * pow(pb_increment, pb) < freq && pb <= 8192) {
      pb++;
    }

    pb--;
    pb += 0x2000;

    printf("key %d (%d:%d) freq %1.4fhz midi %d/%d\n", i, a, b, freq, note, pb);

    keys[i] = (struct key) {
      .ratio_a = a,
      .ratio_b = b,
      .name = names[i],
      .last_send = 0,
      .threshold_min = 150,
      .threshold_max = 800,
      .midinote = note,
      .pitchbend_msb = pb / 128,
      .pitchbend_lsb = pb % 128,
      .channel = CHANNEL_NONE,
      .pan = enable_pan ? ((a * 37 + b * 111 + note * 13) % 64) + 32 : 64
    };
  }

  /* calibration */
  /*
  keys[0].threshold_min = 300;
  keys[0].threshold_max = 500;
  keys[1].threshold_min = 50;
  keys[1].threshold_max = 500;
  keys[22].threshold_min = 100;
  keys[22].threshold_max = 500;
  keys[40].threshold_min = 300;
  keys[42].threshold_min = 80;
  keys[42].threshold_max = 300;
  keys[45].threshold_min = 100;
  keys[45].threshold_max = 600;
  keys[46].threshold_min = 250;
  keys[52].threshold_min = 50;
  keys[52].threshold_max = 200;
  keys[54].threshold_min = 200;
  keys[58].threshold_min =  80;
  keys[58].threshold_max = 600;
  keys[59].threshold_min = 110;
  keys[59].threshold_max = 250;
  keys[60].threshold_min = 200;
  keys[62].threshold_min = 100;
  keys[62].threshold_max = 500;
  keys[64].threshold_min = 100;
  keys[64].threshold_max = 300;
  keys[65].threshold_min = 50;
  keys[65].threshold_max = 300;
  keys[74].threshold_min = 100;
  keys[74].threshold_max = 600;
  keys[78].threshold_min = 500;
  keys[88].threshold_min = 100;
  keys[88].threshold_max = 700;
  keys[89].threshold_min = 100;
  keys[89].threshold_max = 500;
  keys[94].threshold_min =  60;
  keys[94].threshold_max = 400;
  keys[96].threshold_min = 30;
  keys[96].threshold_max = 30+128;
  keys[104].threshold_min = 100;
  keys[104].threshold_max = 700;
  keys[124].threshold_min = 180;
  keys[146].threshold_min = 180;
  */

  /* stuck keys */
  keys[0].threshold_min   = 300; // 1:1
  keys[1].threshold_min   = 190; // 25/96
  keys[2].threshold_min   = 200; // 17:64
  keys[6].threshold_min   = 280; // 9:32
  keys[7].threshold_min   = 200; // 7:24
  keys[9].threshold_min   = 200; // 19:64
  keys[10].threshold_min  = 220; // 3:10:w
  keys[12].threshold_min  = 200; // 21:64
  keys[13].threshold_min  = 200; // 3:1
  keys[15].threshold_min  = 220; // 11:32
  keys[17].threshold_min  = 200; // 45/128
  keys[18].threshold_min  = 200; // 35/96
  keys[20].threshold_min  = 220; // 25:64
  keys[23].threshold_min  = 250; // 105:256
  keys[25].threshold_min  = 280; // 26:64
  keys[27].threshold_min  = 170; // 4/9
  keys[28].threshold_min  = 200; // 9:20
  keys[29].threshold_min  = 180; // 15:32
  keys[39].threshold_min  = 200; // 7:12
  keys[46].threshold_min  = 240; // 27:40
  keys[55].threshold_min  = 120; // 105:128
  keys[59].threshold_min  = 250; // 8:9
  keys[78].threshold_min  = 450; // 27:20
  keys[83].threshold_min  = 200; // 3:2
  keys[97].threshold_min  = 120; // 25:12
  keys[102].threshold_min = 200; // 9:4
  keys[107].threshold_min = 180; // 5:2
  keys[116].threshold_min = 650; // 25:8
  keys[118].threshold_min = 300; // 13:8
  keys[124].threshold_min = 200; // 18:5
  keys[159].threshold_min = 230; // 8:1

  /* quiet keys */
  keys[32].threshold_min  =  50; // 1:2
  keys[42].threshold_min  = 100; // 3:5
  keys[52].threshold_min  =  90; // 25:32
  keys[65].threshold_min  = 120; // 25:24
  keys[96].threshold_min  =  80; // 2:1

  for (i=0; i<ADCS*8; i++) {
    int min = keys[i].threshold_min;
    keys[i].threshold_max =
      min*4 > 1023 ? 1023 : min*4;
  }

  keys[64].threshold_max = 400; // 1:1

  set_minval();
  set_maxval();

  for (i=0; i<ADCS*8; i++) {
    int min = keys[i].threshold_min + 10;
    int max = keys[i].threshold_max;
    if (max < min) {
      printf("suspicious calibration data for key %d; min = %d, max = %d\n", i, min, max);
      max = min;
    }
    keys[i].threshold_min = min;
    keys[i].threshold_max = max;
  }
}

struct channel {
  int in_use;
  int cur_note;
  int volume;
  int disable;
  uint64_t note_begin;
};

struct channel channels[CHANNELS];

void channelinit(int inum) {
  int i;
  for (i=0; i<CHANNELS; i++) {
    channels[i] = (struct channel) {
      .in_use = 0,
      .cur_note = -1,
      .volume = 0,
      .disable = 0
    };

    /* fluidsynth makes the odd channels quiet for some reason */
    //if (i%2 == 1)
    //  channels[i].disable = 1;

    allnoteoff(i);

    //instrument(i, 0, 0, 22); /* proteus flute */
    //instrument(i, 0, 0, 8);   /* proteus synth bass */
    //instrument(i, 0, 0, 47);  /* proteus classical piano */
    //instrument(i, 0, 0, 93);  /* proteus drawbar */
    //instrument(i, 0, 0, 111);   /* proteus organ */
    //instrument(i, 0, 0, 77); /* 77/1 proteus obx saws */
    //instrument(i, 0, 3, 62); /* proteus chamber group */
    //instrument(i, 0, 3, 82); /* proteus real matrix */

    if (inum == -1) {
      instrument_num = 52; /* GM ahhh */
      instrument(i, 0, 0, instrument_num);
    } else {
      instrument(i, 0, 0, inum);
      printf("setting instrument %d\n", inum);
    }
  }
}

int claim_channel() {
  int min_volume = 128;
  int min_volume_index = 0;
  int i;

  for (i=0; i<CHANNELS; i++) {
    printf("(%d,%c,%d) ", channels[i].cur_note, channels[i].in_use ? 'U' : 'F', channels[i].volume);
  }
  printf("\n");

  for (i=0; i<CHANNELS; i++) {
    if (!channels[i].in_use && !channels[i].disable) {
      channels[i].in_use = 1;
      printf("claimed unused channel %d\n", i);
      return i;
    }
  }

  for (i=0; i<CHANNELS; i++) {
    if (channels[i].volume < min_volume && !channels[i].disable) {
      min_volume = channels[i].volume;
      min_volume_index = i;
    }
  }

  printf("claimed in-use channel %d\n", i);
  allnoteoff(min_volume_index);
  return min_volume_index;
}

void release_channel(int i) {
  //printf("released channel %d\n", i);
  channels[i].in_use = 0;
  channels[i].volume = 0;
  channels[i].cur_note = -1;

  allnoteoff(i);
}

int main(int argc, char** argv) {
  int i;
  int chan;
  int ret;
  uint64_t iter = 0;
  int print_freq = 200;
  int calibration_data_freq = 200;
  int vals[ADCS*8];

  int pinstate[2][2] = {{ 0 }};

  int opt;
  int mincal = 0;
  int maxcal = 0;

  while ((opt = getopt (argc, argv, "ab")) != -1) {
    switch(opt) {
      case 'a': mincal = 1; break;
      case 'b': maxcal = 1; break;
      default:
        printf("unexpected option\n");
        exit(1);
    }
  }

  noteinit();

  midifd = open(dev, O_WRONLY, 0);

  if (midifd < 0) {
    printf("can't open device %s\n", dev);
    return 1;
  }

  channelinit(-1);

  ret = wiringPiSetup();
  printf("wiringPiSetup return %d\n", ret);

  pinMode(SREG_CLK, OUTPUT);
  pinMode(SREG_DATA, OUTPUT);
  //pinMode(ADC_CS, OUTPUT);

  pinMode(UP, INPUT);
  pinMode(DOWN, INPUT);

  //wiringPiSPISetup(0, 500000);

  ret = mcp3008Setup(BASE, SPI_CHAN, 2000000);
  printf("mcp3004Setup returned %d\n", ret);

  //adc_cs_low;

  /* clear shift register */
  digitalWrite(SREG_DATA, HIGH);
  for (i=0; i<ADCS; i++) {
    digitalWrite(SREG_CLK, HIGH);
    usleep(TICK);
    digitalWrite(SREG_CLK, LOW);
    usleep(TICK);
  }

  while(1) {
    pinstate[1][0] = digitalRead(UP);
    pinstate[1][1] = digitalRead(DOWN);

    for (i=0; i<2; i++) {
      if (pinstate[0][i] == 0 && pinstate[1][i] == 1) {
        printf("pin %d on\n", i);

        switch (i) {
          case 0: instrument_num--; break;
          case 1: instrument_num++; break;
        }

        if (instrument_num > 255)
          instrument_num = 0;
        if (instrument_num < 0)
          instrument_num = 255;

        channelinit(instrument_num);
      }
      pinstate[0][i] = pinstate[1][i];
    }

    /* read ADCs */
    for (i=0; i<ADCS; i++) {

      if (i==0) {
        digitalWrite(SREG_DATA, LOW);
      }

      digitalWrite(SREG_CLK, HIGH);
      usleep(TICK);
      digitalWrite(SREG_CLK, LOW);
      usleep(TICK);

      if (i==0) {
        digitalWrite(SREG_DATA, HIGH);
        digitalWrite(SREG_CLK, HIGH);
        digitalWrite(SREG_CLK, LOW);
      }

      for (chan=7; chan >= 0; chan--) {
        vals[i*8+chan] = 1023 - analogRead(BASE+chan);
      }
    }
    usleep(1);

    iter++;

    /* send appropriate midi commands */
    for (i=0; i<ADCS*8; i++) {
      int val = vals[i];
      struct key *key = &keys[i];
      double volumef;
      int volume;
      int send_note_on = 0;


      if (key->ratio_a == 0 || key->ratio_b == 0)
        continue;

      if (val < key->minval) {
        key->minval = val;
      }

      if (val > key->maxval) {
        key->maxval = val;
      }

      //printf("key %d val %d min %d channel %d\n", i, val, key->threshold_min, key->channel);

      if (val <= key->threshold_min && key->channel == CHANNEL_NONE)
        continue;

      volumef =
        (double)(val - key->threshold_min) /
        (double)(key->threshold_max - key->threshold_min);

      if (volumef >= 1.0)
        volumef = 1.0;
      else if (volumef < 0)
        volumef = 0.0;
      else
        volumef = pow(volumef, curve_exponent);

      volume = (int)(volumef*127.0);

      //printf("key %d calculated volume %d\n", i, volume);

      if (volume == 0 && key->channel == CHANNEL_NONE)
        continue;

      if (key->channel == CHANNEL_NONE) {
        key->channel = claim_channel();
        pitchbend(key->channel, key->pitchbend_msb, key->pitchbend_lsb);
        //pan(key->channel, key->pan);
        channels[key->channel].note_begin = iter;
        channels[key->channel].cur_note = i;
        send_note_on = 1;
      }

      if (key->channel == CHANNEL_NONE)
        exit(-1);

      /* limit the decay */
      if (channels[key->channel].cur_note == i)
        volume = max(volume, channels[key->channel].volume-2);

      if (volume == 0) {
        printf("volume 0 note %d, key->channel = %d channels[key->channel].cur_note = %d\n",
          i, key->channel, channels[key->channel].cur_note);
        if (channels[key->channel].cur_note == i) {
          note_off(key->channel, key->midinote, 64);
          release_channel(key->channel);
          key->channel = CHANNEL_NONE;
          printf("ended note %d\n", i);
        }

        if (key->channel != CHANNEL_NONE) {
          key->channel = CHANNEL_NONE;
        }
      } else if (send_note_on) {
        chanvol(key->channel, volume);
        note_on(key->channel, key->midinote, 64);
        channels[key->channel].volume = volume;
        printf("key press  %d (%d/%d) volume %d chan %d val %d\n", i, key->ratio_a, key->ratio_b, volume, key->channel, vals[i]);
      } else if (channels[key->channel].cur_note != i) {
        /* channel was stolen */
        printf("channel stolen %d chan %d\n", i, key->channel);
        key->channel = CHANNEL_NONE;
      } else if (volume != channels[key->channel].volume) {
        /* skip some updates to long-held notes to reduce midi command traffic */
        if (iter - channels[key->channel].note_begin < 50 || iter %2 == 0) {
          chanvol(key->channel, volume);
          channels[key->channel].volume = volume;

          printf("key update %d (%d/%d) volume %d chan %d val %d\n", i, key->ratio_a, key->ratio_b, volume, key->channel, vals[i]);
        }
      }
    }

    if (iter % print_freq == 0) {
      for (i=0; i<ADCS; i++) {
        for (chan=0; chan<8; chan++) {
          printf("%10s %4d ", names[i*8+chan], vals[i*8+chan]);
        }
        printf("\n");
      }
      printf("\n");
    }

    if ((iter % calibration_data_freq == 0) && (mincal || maxcal)) {
      char* funcname;
      char* filename;
      char* valname;
      FILE *f;

      printf("opening file\n");

      if (mincal) {
        funcname = "set_minval";
        filename = "min_calibration.h";
        valname = "threshold_min";
      } else if (maxcal) {
        funcname = "set_maxval";
        filename = "max_calibration.h";
        valname = "threshold_max";
      } else {
        printf("something went wrong\n");
        exit(1);
      }

      f = fopen(filename, "w+");
      
      if (!f) {
        printf("can't open calibration file\n");
        exit(1);
      }
 
      printf("writing to file %s\n", filename);
 
      fprintf(f, "void %s() {\n", funcname);

      printf("void %s() {\n", funcname);

      for (i=0; i<ADCS*8; i++) {
        int minval = keys[i].minval;
        int maxval = keys[i].maxval;
        int range = maxval - minval;

        printf("i %d min %d max %d range %d val %s\n", i, minval, maxval, range, valname);
        if (range <= 0) {
          minval = 0;
          maxval = 1023;
          range = 1023;
        }
        fprintf(f, "  keys[%d].%s = %d;\n", i, valname, maxval);
        printf("  keys[%d].%s = %d;\n", i, valname, maxval);
      }
      fprintf(f, "}\n");
      printf("}\n");

      printf("closing file\n");

      fclose(f);
    }
  }

  return 0;
}
